#!/usr/bin/python3
#
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
# Copyright (C) 2020 Cumulus Networks, Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License
# Agreement available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
#
from __future__ import print_function
from enum import IntEnum

import sys
import argparse
import socket
import json

from ctypes import *
from ctypes import sizeof


class ArgParseError(RuntimeError):
    ''' Argument parser error - runtime '''
    pass


class ChassisCtlCmd(IntEnum):
    ''' Chassis CTL command that will be sent via IPC to chassis manager'''
    MAINTENANCE_ENABLE = 1
    MAINTENANCE_DISABLE = 2
    LINECARD_ENABLE = 3
    LINECARD_DISABLE = 4
    SHOW_SUMMARY = 5
    SHOW_DETAIL = 6
    AUTO_PROVISION_ENABLE = 7
    AUTO_PROVISION_DISABLE = 8
    INVALID = 8

class ChassisCtlCmdStatus(IntEnum):
    ''' Chassis CTL command status '''
    SUCCESS = 0
    INVALID_SLOT = 2
    INVALID_CMD = 3
    UNKNOWN_FAILURE = 4

class LineCardTypes(IntEnum):
    '''
    Linecards that are supported on SN4800; Got the following types from SDK
        - 4x400GE
        - 8x200GE
        - 16x100GE
    '''
    SX_MGMT_SLOT_CARD_TYPE_INVALID_E = 0
    SX_MGMT_SLOT_CARD_TYPE_MSN_4_400G_E = 1
    SX_MGMT_SLOT_CARD_TYPE_MSN_8_200G_E = 2
    SX_MGMT_SLOT_CARD_TYPE_MSN_16_100G_E = 3
    SX_MGMT_SLOT_CARD_TYPE_MIN = SX_MGMT_SLOT_CARD_TYPE_INVALID_E
    SX_MGMT_SLOT_CARD_TYPE_MAX = SX_MGMT_SLOT_CARD_TYPE_MSN_16_100G_E

class BaseStructure(Structure):
    '''
    This class is to intialize the ctypes/struct object.
    Any class that dereives from this class can set
    default values for the fields
    '''
    def __init__(self, **kwargs):
        """
        Ctypes.Structure with integrated default values.

        :param kwargs: values different to defaults
        :type kwargs: dict
        """

        values = type(self)._defaults_.copy()
        for (key, val) in kwargs.items():
            values[key] = val

        super().__init__(**values)            # Python 3 syntax

class ChassisCtlIpcCmd(BaseStructure):
    '''
    new IPC command header that will be sent over the wire; the server
    will send this header along with data
        cmd - IPC command type
        ipclen - Length of IPC command
        seq - Sequence number
        slot - Line card slot; It can be 1-8 or 255 (for all slots)
        status - response from server with information on whether the
               command succeeded or not
        type - provisioning type : It can be [16x100GE, 4x400GE]
    '''
    _pack_ = 1
    _fields_ = [("cmd", c_uint32),
                ("ipclen", c_uint32),
                ("seq", c_int32),
                ("slot", c_uint32),
                ("status", c_uint32),
                ("type", c_uint32),
               ]
    _defaults_ = {"cmd": ChassisCtlCmd.INVALID,
                  "ipclen": 0,
                  "seq": 0,
                  "slot": 0,
                  "status": 0,
                  "type": 0,
                 }

    def __new__(cls, in_cmd=None):
        if in_cmd is not None:
            return cls.from_buffer_copy(in_cmd)

        return Structure.__new__(cls)


class ChassisCtlIpcResp(BaseStructure):
    '''
    IPC response format that will be sent over the wire from
    chassis manager
    '''
    _pack_ = 1
    _fields_ = [("serial", c_char * 32),
                ("card_type", c_char * 32),
                ("ini_major", c_uint32),
                ("ini_minor", c_uint32),
                ("state", c_char * 32),
                ("provisioned", c_uint32),
                ("uptime", c_uint64),
               ]
    _defaults_ = {"serial": b'',
                  "card_type": b'',
                  "ini_major": 0,
                  "ini_minor": 0,
                  "state": b'',
                  "provisioned": 0,
                  "uptime" : 0
                 }

    def __new__(cls, in_resp=None):
        if in_resp is not None:
            return cls.from_buffer_copy(in_resp)

        return Structure.__new__(cls)


class ChassisCtlIpcFailResp(BaseStructure):
    '''
    IPC resonse format when the command fails. The IPC command
    header will have status field which will indicate if the command
    succeeded/failed
    '''
    _pack_ = 1
    _fields_ = [("status", c_char * 128)]
    _defaults_ = {"status": b''}

    def __new__(cls, in_resp=None):
        if in_resp is not None:
            return cls.from_buffer_copy(in_resp)

        return Structure.__new__(cls)


def set_slot_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the slot
    appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'slot':
        if value == 'all':
            msg.slot = int(255)
        else:
            msg.slot = int(value)


def set_show_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the cmd
    appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'show':
        if value == 'summary':
            msg.cmd = ChassisCtlCmd.SHOW_SUMMARY
        elif value == 'detail':
            msg.cmd = ChassisCtlCmd.SHOW_DETAIL
        elif value == None:
            return None

    return 'success'


def set_provision_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the provision
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'provision':
        if value == '16x100GE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_ENABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_MSN_16_100G_E
        elif value == '4x400GE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_ENABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_MSN_4_400G_E
        elif value == '8x200GE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_ENABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_MSN_8_200G_E
        elif value == 'DISABLE':
            msg.cmd = ChassisCtlCmd.AUTO_PROVISION_DISABLE
            msg.type = LineCardTypes.SX_MGMT_SLOT_CARD_TYPE_INVALID_E
        else:
            return None

    return 'success'

def set_linecard_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the linecard
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'linecard':
        if value == 'enable':
            msg.cmd = ChassisCtlCmd.LINECARD_ENABLE
        elif value == 'disable':
            msg.cmd = ChassisCtlCmd.LINECARD_DISABLE
        else:
            return None

    return 'success'


def set_maintenance_msg(key, value, msg):
    '''
    Given key and value from the parse dictionary, set the maintenance
    field appropriately in the CTL message; This API is called to build
    the CTL message
    '''
    if key == 'maintenance':
        if value == 'enable':
            msg.cmd = ChassisCtlCmd.MAINTENANCE_ENABLE
        elif value == 'disable':
            msg.cmd = ChassisCtlCmd.MAINTENANCE_DISABLE
        else:
            return None

    return 'success'


def build_ctl_msg(cmddict, seq):
    '''
    Given a dictionary of parsed cmdline, build CTL IPC message
    '''
    msg = ChassisCtlIpcCmd()
    msg.ipclen = sizeof(ChassisCtlIpcCmd)
    msg.seq = seq
    for key, value in cmddict.items():
        set_slot_msg(key, value, msg)

        if set_show_msg(key, value, msg) == None:
            continue

        if set_linecard_msg(key, value, msg) == None:
            continue

        if set_provision_msg(key, value, msg) == None:
            continue

        if set_maintenance_msg(key, value, msg) == None:
            continue

    return msg


def parse_cmdline(argv):
    ''' Parse the command line '''
    parser = argparse.ArgumentParser(
        description='CTL interface to chassis manager')

    parser.add_argument('-l', '--slot',
                        dest="slot",
                        required=True,
                        action='store',
                        help='specify the linecard slot (\'all\' or 1-8).')
    parser.add_argument('-m', '--maintenance',
                        dest="maintenance",
                        required=False,
                        action='store',
                        help='maintenance mode enable/disable',
                        choices=['enable', 'disable'],
                        default='disable')
    parser.add_argument('-e', '--lcenable',
                        dest='linecard',
                        required=False,
                        action='store',
                        help='linecard dwmode enable/disable',
                        choices=['enable', 'disable'],
                        default='none')
    parser.add_argument('-f', '--gb-fw-update',
                        dest="gb-fw-path",
                        required=False,
                        action='store',
                        help='update linecard gearbox fw',
                        default=None)
    parser.add_argument('-p', '--gb-ini-update',
                        dest="gb-ini-path",
                        required=False,
                        action='store',
                        help='update linecard gearbox ini',
                        default=None)
    parser.add_argument('-s', '--show',
                        dest="show",
                        required=False,
                        action='store',
                        help='show detail/summary',
                        choices=['summary', 'detail'],
                        default=None)
    parser.add_argument('-c', '--provision',
                        dest="provision",
                        required=False,
                        action='store',
                        help='provision line card - provide type',
                        choices=['16x100GE', '4x400GE', 'DISABLE'],
                        default=None)
    pretty_print = False
    parser.add_argument('-j', '--json',
                        dest="json",
                        required=False,
                        action='store_const',
                        help='display json output',
                        default=pretty_print,
                        const=not pretty_print)

    try:
        if len(argv) == 1:
            parser.print_help(sys.stderr)
            sys.exit(1)

        args = parser.parse_args(argv)
        mydict = vars(args)
        #print(mydict)
        return mydict
    except ArgParseError:
        exit(-1)


def send_ctl_toserver(cmd, cmddict):
    ''' connect to server and do IPC '''
    sock = socket.socket(socket.AF_UNIX,
                         socket.SOCK_STREAM)
    #print('connecting to server', file=sys.stderr)
    try:
        serveraddr = '/tmp/chassismgr'
        sock.connect(serveraddr)
        sock.sendall(cmd)

        data, sender = sock.recvfrom(1024)
        #print('received ', len(data), file=sys.stderr)
        parse_ctl_response(cmd, data, cmddict)

    except socket.error as err:
        print('failed to connect', err, file=sys.stderr)
        sys.exit(1)

    finally:
        #print('closing socket', file=sys.stderr)
        sock.close()

def print_failure_bail(payload):
    '''
    We received IPC command failure; We need to decode the failure
    message in the response
    '''
    class Message(Structure):
        ''' a dynamic structure to unpack contents on the wire '''
        _pack_ = 1
        _fields_ = [("header", ChassisCtlIpcCmd),
                    ("resp", ChassisCtlIpcFailResp)
                   ]

        def __new__(cls, obj=None):
            if obj is not None:
                return cls.from_buffer_copy(obj)

            return Structure.__new__(cls)

    response = Message.from_buffer_copy(payload)
    print(response.resp.status.decode('utf-8'))
    return response


def get_ipcresponse_cnt(cmd, payload):
    '''
    Given the payload, find the number of IPC responses we received
    '''
    if sizeof(ChassisCtlIpcCmd) > len(payload):
        raise ValueError(-1, "Buffer smaller than header")

    header = ChassisCtlIpcCmd.from_buffer_copy(payload)
    ''' validate the response '''
    if validate_response(cmd, header) == False:
        print_failure_bail(payload)
        return int(0)

    if header.slot == 255:
        response_cnt = int(8)
    else:
        response_cnt = int(1)

    if response_cnt > len(payload):
        raise ValueError(-1, "Buffer smaller than header and data")

    return response_cnt


def validate_response(cmd, header):
    '''
    Validate the IPC command and the response
    '''
    if cmd.cmd != header.cmd:
        return False

    if header.status != ChassisCtlCmdStatus.SUCCESS:
        print('Validation failed with status %d' % header.status)
        return False

    return True


def parse_ctl_response(cmd, payload, cmddict):
    ''' API to parse the payload received on the wire'''
    response_cnt = get_ipcresponse_cnt(cmd, payload)
    #print("printing the response cnt = %d" % response_cnt)

    class Message(Structure):
        ''' a dynamic structure to unpack contents on the wire '''
        _pack_ = 1
        _fields_ = [("header", ChassisCtlIpcCmd),
                    ("response", ChassisCtlIpcResp * response_cnt)
                   ]

        def __new__(cls, obj=None):
            if obj is not None:
                return cls.from_buffer_copy(obj)

            return Structure.__new__(cls)

    message = Message.from_buffer_copy(payload)

    print_ctl(message, response_cnt, cmddict)
    return message


def print_ctl(message, response_cnt, cmddict):
    ''' print CTL output on the console '''
    if message.header.cmd == ChassisCtlCmd.SHOW_SUMMARY:
        print_ctl_summary(message, response_cnt, cmddict)
    elif message.header.cmd == ChassisCtlCmd.SHOW_DETAIL:
        print_ctl_detail(message, response_cnt, cmddict)

def print_json_summary(message, response_cnt, cmddict):
    '''API to print the summary in JSON format '''
    data = {}
    decoding = 'utf-8'
    for i in range(0, response_cnt):
        key = 'linecard' + str(i)
        data[key] = {}
        data[key]['slot'] = str(i)
        data[key]['card_type'] = message.response[i].card_type.decode(decoding)
        data[key]['serial'] = message.response[i].serial.decode(decoding)
        data[key]['ini_major'] = message.response[i].ini_major
        data[key]['ini_minor'] = message.response[i].ini_minor
        data[key]['state'] = message.response[i].state.decode(decoding)

    json_data = json.dumps(data, indent=4)
    print(json_data)

def print_ctl_summary(message, response_cnt, cmddict):
    ''' print the CTL Summary on console '''
    if cmddict['json'] == True:
        print_json_summary(message, response_cnt, cmddict)
        return

    decoding = 'utf-8'
    print("-" * 50)
    print("{:<3} {:<18} {:<15}".format('Slot', 'Card type',
                                 'State'))
    print("-" * 50)
    for i in range(0, response_cnt):
        print("{:<3} {:<18} {:<15}".format(str(i),
            message.response[i].card_type.decode(decoding),
            message.response[i].state.decode(decoding)))

        if response_cnt is 1:
            break



def print_ctl_detail(message, response_cnt, cmddict):
    ''' print the CTL Detail '''
    if cmddict['json'] == True:
        print_json_summary(message, response_cnt, cmddict)
        return

    decoding = 'utf-8'
    print("-" * 90)
    print("{:<3} {:<18} {:<15} {:<20} {:<20} {:<15}".format('Slot',
                                                            'Card type',
                                                            'Serial',
                                                            'INI major',
                                                            'INI minor',
                                                            'State'))
    print("-" * 90)
    for i in range(0, 8):
        print("{:<3} {:<18} {:<15} {:^20} {:^20} {:<15}".format(
            str(i),
            message.response[i].card_type.decode(decoding),
            message.response[i].serial.decode(decoding),
            message.response[i].ini_major,
            message.response[i].ini_minor,
            message.response[i].state.decode(decoding)))

        if response_cnt is 1:
            break


# ----------------------------------------------------------
#
#                         e n t r y
#
# ----------------------------------------------------------
if __name__ == "__main__":
    CMDDICT = parse_cmdline(sys.argv[1:])
    CTLMSG = build_ctl_msg(CMDDICT, 1000)
    RESPONSE = send_ctl_toserver(CTLMSG, CMDDICT)
