#!/usr/bin/env python
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
# Copyright 2016-2020 Cumulus Networks, Inc
#
# ecmpcalc utility helpers

import subprocess
import cumulus.porttab
import re
import ipaddr
import dpkt
import socket

class ParseError(RuntimeError):
    pass


####################################################################
# ecmpcalc IP Frame Field Objects
####################################################################
class FrameField:
    '''
    Represents a 16-bit field from an IP frame.
    '''
    name = 'Unknown Frame Field'
    def __init__(self, num):
        self.value = num

    @classmethod
    def from_int(self, num):
        return self(num)

    @classmethod
    def from_string(self, string):
        try:
            value = int(string, 0)
        except ValueError:
            raise ParseError('%s is not an int for frame field %s' %
                             (string, self.name))
        return self(value)

    def __str__(self):
        return self.name + ':' + str(self.value)
    __repr__ = __str__

    def asInt(self):
        return self.value

class ProtocolID(FrameField):
    '''
    IPV4 Protocol ID
    '''
    name = 'Protocol ID'
    protocolMap = { 'icmp' : 1,
                    'tcp'  : 6,
                    'udp'  : 17,
                    'icmp6': 58,
                  }

    @property
    def isTCPorUDP(self):
        return self.asInt() in (6, 17)

    @classmethod
    def from_string(self, string):
        if string in self.protocolMap.keys():
            return self(self.protocolMap[string])

        return self.from_int(int(string))

class L4Port(FrameField):
    name = 'L4 Port'
class VLANID(FrameField):
    name = 'VLAN ID'
class HardwarePort(FrameField):
    name = 'Hardware Port'
class HardwareModule(FrameField):
    name = 'Hardware Module'
class HashSeed(FrameField):
    name = 'Hash Seed'
class Ip6FlowLabel(FrameField):
    name = 'Ip6 flow label'

class Frame:
    '''
    Represents an IP frame with Trident-specific attributes
    '''
    field_desc = { 'src'      : 'Source IP Address',
                   'dst'      : 'Destination IP Address',
                   'vid'      : 'VLAN ID',
                   'sport'    : 'Source Layer 4 Port',
                   'dport'    : 'Destination Layer 4 Port',
                   'protocol' : 'Layer 3 Protocol',
                   'sportid'  : 'Source Hardware Port ID',
                   'smodid'   : 'Source Hardware Module ID',
                   'dportid'  : 'Destination Hardware Port ID',
                   'dmodid'   : 'Destination Hardware Module ID',
                   'ip6_flow_label' : 'Ipv6 flow label'
                   }
    def __init__(self, frameInfo):
        self.src = frameInfo["src"] # IPV4 source
        self.dst = frameInfo["dst"] # IPV4 destination
        self.vid = frameInfo["vid"] # vlan ID
        self.sport = frameInfo["sport"] # L4 source port
        self.dport = frameInfo["dport"] # L4 destination port
        self.protocol = frameInfo["protocol"] # IP protocol
        self.sportid = frameInfo["sportid"] # hardware source port
        self.smodid = frameInfo["smodid"] # hardware source module
        self.dportid = frameInfo["dportid"] # hardware destination port
        self.dmodid = frameInfo["dmodid"] # hardware destination module
        self.ip6_flow_label = frameInfo["ip6_flow_label"]

    # properties to return the high (most significant 16-bits) and low (least
    # significant 16-bits) of the frame's IPV4 address

    @property
    def srcH(self):
        if self.src is None:
            return None
        return self.src.high()
    @property
    def srcL(self):
        if self.src is None:
            return None
        return self.src.low()
    @property
    def dstH(self):
        if self.dst is None:
            return None
        return self.dst.high()
    @property
    def dstL(self):
        if self.dst is None:
            return None
        return self.dst.low()

    def __str__(self):
        validFields = []
        for fieldName in ('src', 'dst', 'vid', 'sport', 'dport', 'protocol',
                          'sportid', 'smodid', 'dportid', 'dmodid','ip6_flow_label'):
            fieldValue = getattr(self, fieldName)
            if fieldValue is not None:
                validFields.append(fieldName + ':(' + str(fieldValue) + ')')

        return 'IP Frame: ' + ' '.join(validFields)
    __repr__ = __str__

    def _buildEnetHdr(self, smac, dmac, vlan, version):
        # build enet hdr manually - we need to always tag the packet and
        # dpkt currently only constructs untagged enet hdrs
        data = dmac.replace(":", "") + smac.replace(":", "")

        if vlan:
            data += "%04x" % dpkt.ethernet.ETH_TYPE_8021Q
            data += "%04x" % vlan

        if version is 4:
            data += "%04x" % dpkt.ethernet.ETH_TYPE_IP
        else:
            data += "%04x" % dpkt.ethernet.ETH_TYPE_IP6

        return data

    def _buildv6Packet(self, ehdr, ip):
        # build the v6 bytes manually as dpkt module
        # has issues with v6 header
        p = bytes(ip.data)
        s = dpkt.struct.pack('>16s16sxBH', ip.src, ip.dst, ip.nxt, len(p))
        s = dpkt.in_cksum_add(0, s)
        s = dpkt.in_cksum_add(s, p)
        try:
            ip.data.sum = dpkt.in_cksum_done(s)
        except AttributeError:
            pass
        v6 = ip.pack_hdr() + bytes(ip.data)
        pkt = ehdr + v6.encode('hex')
        return pkt

    def buildVisibilityPkt(self, smac, dmac, vid):
        # ip payload
        if self.protocol.value == dpkt.ip.IP_PROTO_UDP:
            ipl = dpkt.udp.UDP(sport=self.sport.value, dport=self.dport.value)
        elif self.protocol.value == dpkt.ip.IP_PROTO_TCP:
            ipl = dpkt.tcp.TCP(sport=self.sport.value, dport=self.dport.value)
        else:
            return None

        # ip pkt
        dstAddr = ipaddr.IPAddress(self.dst.addrstr)
        srcAddr = ipaddr.IPAddress(self.src.addrstr)
        if dstAddr.version is 4:
            ip = dpkt.ip.IP(
                    src=socket.inet_pton(socket.AF_INET, str(srcAddr)),
                    dst=socket.inet_pton(socket.AF_INET, str(dstAddr)),
                    p=self.protocol.value, data=ipl)
            ip.len += len(ipl)
        elif dstAddr.version is 6:
            ip = dpkt.ip6.IP6(
                    src=socket.inet_pton(socket.AF_INET6, str(srcAddr)),
                    dst=socket.inet_pton(socket.AF_INET6, str(dstAddr)),
                    nxt=self.protocol.value, data=ipl, plen=len(ipl),
                    flow=self.ip6_flow_label.asInt(),
                    hlim=255)
        else:
            return None

        # ethernet pkt
        ehdr = self._buildEnetHdr(smac, dmac, vid, dstAddr.version)
        if dstAddr.version is 6:
            pkt = self._buildv6Packet(ehdr, ip)
        else:
            pkt = ehdr + str(ip).encode("hex")

        return pkt

    def visibilityPktMissingFields(self):
        missing = set()
        for fieldName in ('src', 'dst', 'sport', 'dport', 'protocol'):
            fieldValue = getattr(self, fieldName)
            if fieldValue is None:
                missing.add(self.field_desc[fieldName])
        return missing

########################################################################
# IP and Ethernet Addresses
########################################################################
class Address(FrameField):
    def __init__(self, string):
        self.addrstr = string

        if self.validate() is not True:
           raise ParseError('"%s" is not a %s' % (self.addrstr, self.name))

    @classmethod
    def from_string(self, string):
        return self(string)

    def __str__(self):
        return self.name + ':' + self.addrstr
    __repr__ = __str__

    def _checkchars(self, valid):
        for char in self.addrstr:
           if char not in valid:
               raise ParseError('invalid character "%s" in %s: "%s"' %
                                (char, self.name, self.addrstr))
        return True

    def validate(self):
        raise NotImplementedError

    def asInt(self):
        raise NotImplementedError

class MACAddress(Address):
    name = 'MAC Address'
    def validate(self):
        split = self.addrstr.lower().split(':')
        if len(split) != 6:
            return False
        return self._checkchars(string.hexdigits + ':')

class IPAddress(Address):
    name = 'IP Address'

    def __init__(self, string):
        Address.__init__(self, string)
        self.addr = ipaddr.IPAddress(string)

    def validate(self):
        try:
            addr = ipaddr.IPAddress(self.addrstr)
        except ValueError, e:
            raise ParseError(e)

        return True

    def _fold6(self, addrint):
        # IPv6 addresses are "folded" into 32bits: IP[127:96] ^ IP[96:64] ^
        # IP[63:32] ^ IP[31:0]. Some BCM parts support other methods of
        # folding, but this is the default.
        return (((addrint >> 96) & 0xffffffff) ^
                ((addrint >> 64) & 0xffffffff) ^
                ((addrint >> 32) & 0xffffffff) ^
                ((addrint >> 0) & 0xffffffff))

    def asInt(self):
        if self.addr.version is 4:
            return int(self.addr)
        elif self.addr.version is 6:
            return self._fold6(int(self.addr))
        else:
            raise RuntimeError('%s is not a IPv4 or IPv6 address' % str(self.addr))

    def high(self):
        return IPAddressHigh(self.addrstr)

    def low(self):
        return IPAddressLow(self.addrstr)

class IPAddressHigh(IPAddress):
    name = 'IP Address (High 16-bits)'
    def asInt(self):
        val = IPAddress.asInt(self)
        return (val & 0xffff0000) >> 16

class IPAddressLow(IPAddress):
    name = 'IP Address (Low 16-bits)'
    def asInt(self):
        val = IPAddress.asInt(self)
        return val & 0xffff


